/* AP26-boinc.h --

   Modified for SSE2 support 27 September 2009 by Bryan Little

   This file is a replacement for AP26.h in Jaroslaw Wroblewski's AP26
   sample implementation. Most of my changes are between #if REPORT_FREQUENCY
   #endif clauses, they exist only to allow the program to checkpoint part
   way through the SearchAP26(K,SHIFT) routine. Other minor changes are to
   use int64_t instead of long long for compilers that don't conform to C99.

   See http://www.geocities.com/g_w_reynolds/AP26/AP26-boinc-src.zip for the
   latest version of this file.

   See http://www.math.uni.wroc.pl/~jwr/AP26/AP26.zip for the sample
   implementation containing the rest of the source required.

   See http://www.math.uni.wroc.pl/~jwr/AP26/AP26v3.pdf for information
   about how the algorithm works and for the copyleft notice.
*/

#include "CONST-boinc.H"


// function to select elements from two vectors based on a selection mask
__m128i vec_sel( __m128i a, __m128i b, __m128i mask ){

        b = _mm_and_si128( b, mask );
        a = _mm_andnot_si128( mask, a );

        return _mm_or_si128( a, b );
}


#if (REPORT_FREQUENCY > 1)
void SearchAP26(int K, int SHIFT, int START_ITER)
#else
void SearchAP26(int K, int SHIFT)
#endif
{ 
int64_t n, n43, n47, n53, n59;
int i43, i47, i53, i59;
int i3, i5, i31, i37, i41;

int j, jj;

int64_t STEP;
int64_t n0;
int64_t sito;
int64_t S31, S37, S41, S43, S47, S53, S59;

#if (REPORT_FREQUENCY==85 || REPORT_FREQUENCY==1355)
int iter = 0;
#endif

STEP=K*PRIM23;
n0=(N0*(K%17835)+((N0*17835)%MOD)*(K/17835)+N30)%MOD;

S31=(PRES2*(K%17835)+((PRES2*17835)%MOD)*(K/17835))%MOD;
S37=(PRES3*(K%17835)+((PRES3*17835)%MOD)*(K/17835))%MOD;
S41=(PRES4*(K%17835)+((PRES4*17835)%MOD)*(K/17835))%MOD;
S43=(PRES5*(K%17835)+((PRES5*17835)%MOD)*(K/17835))%MOD;
S47=(PRES6*(K%17835)+((PRES6*17835)%MOD)*(K/17835))%MOD;
S53=(PRES7*(K%17835)+((PRES7*17835)%MOD)*(K/17835))%MOD;
S59=(PRES8*(K%17835)+((PRES8*17835)%MOD)*(K/17835))%MOD;

//quick loop vectors
__m128i svec1		= _mm_set_epi16(S59%61, S59%67, S59%71, S59%73, S59%79, S59%83, S59%89, S59%97);
__m128i svec2		= _mm_set_epi16(S59%101, S59%103, S59%107, S59%109, S59%113, S59%127, S59%131, S59%137);

__m128i mvec1		= _mm_set_epi16(MOD%61, MOD%67, MOD%71, MOD%73, MOD%79, MOD%83, MOD%89, MOD%97);
__m128i mvec2		= _mm_set_epi16(MOD%101, MOD%103, MOD%107, MOD%109, MOD%113, MOD%127, MOD%131, MOD%137);

__m128i numvec1_1	= _mm_set_epi16(61, 67, 71, 73, 79, 83, 89, 97);
__m128i numvec2_1	= _mm_set_epi16(60, 66, 70, 72, 78, 82, 88, 96);

__m128i numvec1_2	= _mm_set_epi16(101, 103, 107, 109, 113, 127, 131, 137);
__m128i numvec2_2	= _mm_set_epi16(100, 102, 106, 108, 112, 126, 130, 136);

__m128i zerovec	= _mm_set_epi16(0, 0, 0, 0, 0, 0, 0, 0);
            
       
#include "MAKES-boinc.H"

#if (REPORT_FREQUENCY==7)
for(i31=START_ITER;i31<7;++i31)
{
  /* There are 7 combinations of i31 that reach this block */
  checkpoint(SHIFT,K,i31,0);

#else
for(i31=0;i31<7;++i31)
#endif

for(i37=0;i37<13;++i37)
if(i37-i31<=10&&i31-i37<=4)
#if (REPORT_FREQUENCY==85)
{
  /* There are 85 combinations of i31,i37 that reach this block */
  if (iter++ < START_ITER)
    continue;
  checkpoint(SHIFT,K,iter-1,0);

#endif

for(i41=0;i41<17;++i41)
if(i41-i31<=14&&i41-i37<=14&&i31-i41<=4&&i37-i41<=10)
#if (REPORT_FREQUENCY==1355)
{
  /* There are 1355 combinations of i31,i37,i41 that reach this block */
  if (iter++ < START_ITER)
    continue;
  checkpoint(SHIFT,K,iter-1,0);

#endif

for(i3=0;i3<2;++i3)
for(i5=0;i5<4;++i5)
{ 




n43=(n0+i3*S3+i5*S5+i31*S31+i37*S37+i41*S41)%MOD;
{

for(i43=(PRIME5-24);i43>0;i43--)
{

n47=n43;
for(i47=(PRIME6-24);i47>0;i47--)
{

n53=n47;
for(i53=(PRIME7-24);i53>0;i53--)
{

n59=n53;

// Bryan Little, new fast vectors for loop
// elements in vector arrays are stored IN REVERSE ORDER in SSE2!
__m128i r_numvec1 = _mm_set_epi16(REM(n59,61,6), REM(n59,67,7), REM(n59,71,7), REM(n59,73,7), REM(n59,79,7), REM(n59,83,7), REM(n59,89,7), REM(n59,97,7));
__m128i r_numvec2 = _mm_set_epi16(REM(n59,101,7), REM(n59,103,7), REM(n59,107,7), REM(n59,109,7), REM(n59,113,7), REM(n59,127,7), REM(n59,131,8), REM(n59,137,8));

for(i59=(PRIME8-24);i59>0;i59--)
{
#include "SITO-boinc.H"
{
int b;

for(b=0;b<64;b++)
if((sito>>b)&1)
{
n=n59+(b+SHIFT)*MOD;

if(n%7)
if(n%11)
if(n%13)
if(n%17)
if(n%19)
if(n%23)
#include "IF.H"
{
int64_t m;
int k;
k=0; m=n+STEP*5;
while(PrimeQ(m)){
k++;
m+=STEP;
};


if(k>=10){
m=n+STEP*4;
while(m>0&&PrimeQ(m)){
k++;
m-=STEP;
};
};



if(k>=10)
{
ReportSolution(k,K,m+STEP);
};
};

}

}

// Bryan Little - new fast vector math
if(i59>1){
	n59+=S59;

	__m128i subvec;

	//vector 1
	r_numvec1 = _mm_add_epi16(r_numvec1, svec1);
	//vector 2
	r_numvec2 = _mm_add_epi16(r_numvec2, svec2);

	if(n59>=MOD){
		n59-=MOD;

		__m128i addvec;

		//vector 1
		r_numvec1 = _mm_sub_epi16(r_numvec1, mvec1);
		addvec = _mm_add_epi16(r_numvec1, numvec1_1);
		r_numvec1 = vec_sel( r_numvec1, addvec, _mm_cmpgt_epi16( zerovec, r_numvec1 ) );
		//vector 2
		r_numvec2 = _mm_sub_epi16(r_numvec2, mvec2);
		addvec = _mm_add_epi16(r_numvec2, numvec1_2);
		r_numvec2 = vec_sel(r_numvec2, addvec, _mm_cmpgt_epi16( zerovec, r_numvec2 ) );
	}
	//vector 1
	subvec = _mm_sub_epi16(r_numvec1, numvec1_1);
	r_numvec1 = vec_sel(r_numvec1, subvec, _mm_cmpgt_epi16(r_numvec1, numvec2_1) );
	//vector 2
	subvec = _mm_sub_epi16(r_numvec2, numvec1_2);
	r_numvec2 = vec_sel(r_numvec2, subvec, _mm_cmpgt_epi16(r_numvec2, numvec2_2) );
} //end new fast vector math
           
}               

n53+=S53;
if(n53>=MOD)n53-=MOD;
}

n47+=S47;
if(n47>=MOD)n47-=MOD;
}

n43+=S43;
if(n43>=MOD)n43-=MOD;
}
};


}
#if (REPORT_FREQUENCY==1355)
}
#endif
#if (REPORT_FREQUENCY==85)
}
#endif
#if (REPORT_FREQUENCY==7)
}
#endif
#if 0
}
#endif
#if ASM
asm_fini();
#endif
}
